#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usb.h>
#include "ks0108_usb.h"

static usb_dev_handle *handle;

static unsigned char framebuffer_cur[WIDTH][PHEIGHT] = {};	/*! */

static unsigned char framebuffer_old[WIDTH][PHEIGHT] = {};	/*! */

static int
usbGetStringAscii (usb_dev_handle * dev, int index, int langid, char *buf,
		   int buflen)
{
  char buffer[256];
  int rval, i;

  if ((rval =
       usb_control_msg (dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR,
			(USB_DT_STRING << 8) + index, langid, buffer,
			sizeof (buffer), 1000)) < 0)
    return rval;
  if (buffer[1] != USB_DT_STRING)
    return 0;
  if ((unsigned char) buffer[0] < rval)
    rval = (unsigned char) buffer[0];
  rval /= 2;
  /* lossy conversion to ISO Latin1 */
  for (i = 1; i < rval; i++)
    {
      if (i > buflen)		/* destination buffer overflow */
	break;
      buf[i - 1] = buffer[2 * i];
      if (buffer[2 * i + 1] != 0)	/* outside of ISO Latin1 range */
	buf[i - 1] = '?';
    }
  buf[i - 1] = 0;
  return i - 1;
}


static usb_dev_handle *
findDevice (void)
{
  struct usb_bus *bus;
  struct usb_device *dev;
  usb_dev_handle *handle = 0;

  usb_find_busses ();
  usb_find_devices ();
  for (bus = usb_busses; bus; bus = bus->next)
    {
      for (dev = bus->devices; dev; dev = dev->next)
	{
	  if (dev->descriptor.idVendor == USBDEV_SHARED_VENDOR
	      && dev->descriptor.idProduct == USBDEV_SHARED_PRODUCT)
	    {
	      char string[256];
	      int len;
	      handle = usb_open (dev);
	      if (!handle)
		{
		  fprintf (stderr, "Warning: cannot open USB device: %s\n",
			   usb_strerror ());
		  continue;
		}
	      len =
		usbGetStringAscii (handle, dev->descriptor.iManufacturer,
				   0x0409, string, sizeof (string));
	      if (len < 0)
		{
		  fprintf (stderr,
			   "warning: cannot query manufacturer for device: %s\n",
			   usb_strerror ());
		  goto skipDevice;
		}
	      fprintf (stderr, "seen device from vendor ->%s<-\n", string);
	      if (strcmp (string, "Fhunter") != 0)
		goto skipDevice;
	      len =
		usbGetStringAscii (handle, dev->descriptor.iProduct, 0x0409,
				   string, sizeof (string));
	      if (len < 0)
		{
		  fprintf (stderr,
			   "warning: cannot query product for device: %s\n",
			   usb_strerror ());
		  goto skipDevice;
		}
	      fprintf (stderr, "seen product ->%s<-\n", string);
	      if (strcmp (string, "HID LCD") == 0)
		break;
	    skipDevice:
	      usb_close (handle);
	      handle = NULL;
	    }
	}
      if (handle)
	break;
    }
  if (!handle)
    fprintf (stderr, "Could not find USB device Fhunter/HID LCD\n");
  return handle;
}

static int
usb_transmit (usb_dev_handle * usbhandle, unsigned char receive,
	      unsigned char functionid, unsigned char send[4],
	      unsigned char *buffer, int buffersize)
{
  int nbytes;
  nbytes = usb_control_msg (usbhandle,
			    USB_TYPE_VENDOR | USB_RECIP_DEVICE | (receive <<
								  7),
			    functionid, (send[1] << 8) | send[0],
			    (send[3] << 8) | send[2], buffer, buffersize,
			    5000);
  if (nbytes < 0)
    fprintf (stderr, "error: transmit: %s\n", usb_strerror ());
  return nbytes;

};

static void
send_data_to_device (usb_dev_handle * handle, unsigned char *buffer,
		     unsigned char length)
{
/*-----------------*/
  int n;
  unsigned char cmd[4];
  int wbytes = length+2;
  int blocksize;
  int function;

  function = PSCMD_LCD;
  while (wbytes) {
    blocksize = wbytes;
    wbytes = 0;
    cmd[0] = 0;
    cmd[1] = 0;
    cmd[2] = 0;
    cmd[3] = 0;
    n = usb_transmit(handle,0, PSCMD_LCD, cmd, buffer, blocksize);
    if (n != blocksize) {
      fprintf(stderr, "error: wrong count at writing %x\n",n);
      exit(1);
    }
  }
  return;
};


static void
update_framebuffer (void)
{				/*!       */
  int i, j;
  for (i = 0; i < WIDTH; i++)
    {
      for (j = 0; j < PHEIGHT; j++)
	{
	  framebuffer_old[i][j] = framebuffer_cur[i][j];
	};
    };
};


static void force_full_update(){
	int i,j;
	for(j=0;j<WIDTH;j++) for(i=0;i<PHEIGHT;i++) framebuffer_old[j][i]=~framebuffer_cur[j][i];
};

void
update_screen ()
{
  int i, j, k, length;
  unsigned char buffer[MAX_PACKET];
  for (j = 0; j < PHEIGHT; j++)
    {
      for (i = 0; i < WIDTH; i++)
	{
	  if (framebuffer_cur[i][j] != framebuffer_old[i][j])
	    {
	      length = 3;
	      buffer[0] = i;
	      buffer[1] = j;
	      buffer[2] = framebuffer_cur[i][j];
	      k = 0;
	      while ((framebuffer_cur[i + k][j] != framebuffer_old[i + k][j])
		     && (k < (MAX_PACKET-2)) && ((i + k) < WIDTH))
		{
		  k++;
		  buffer[length] = framebuffer_cur[i + k][j];
		  length++;
		};
	      send_data_to_device (handle, buffer, length-1);
	      i += k;
	    };
	};
    };
  update_framebuffer ();
};

void putpixel(unsigned char x,unsigned char y,unsigned char value){
	unsigned char mask;
	if(x<WIDTH){
		if(y<HEIGHT){
			mask=(1<<(y%8));
			if(value){
				framebuffer_cur[x][y>>3]=(framebuffer_cur[x][y>>3])|mask;
			}else{
				framebuffer_cur[x][y>>3]=(framebuffer_cur[x][y>>3])&(~mask);
			};
		};
	};
};

void clearscreen(unsigned char value){
	int i,j;
	for(i=0;i<WIDTH;i++){
		for(j=0;j<PHEIGHT;j++){
			framebuffer_cur[i][j]=value;
		};
	};
	force_full_update();
};

int open_lcd(){
  usb_init ();
  if ((handle = findDevice ()) == NULL)
    {
      fprintf (stderr,
	       "Could not find USB device \"HID LCD\" with vid=0x%x pid=0x%x\n",
	       USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
	return -1;
    }
    return 0;
};

int close_lcd(){
  usb_close (handle);
  return 0;
};

int read_keys(){
	unsigned char buffer[8];
	unsigned char cmd[4];
	usb_transmit(handle,1,PSCMD_GET,cmd,buffer,sizeof(buffer));
	return buffer[0];
};

void set_led(unsigned char status){
	unsigned char buffer[8];
	unsigned char cmd[4];
	cmd[0]=status;
	usb_transmit(handle,1,PSCMD_LED,cmd,buffer,sizeof(buffer));
	return;
};
